﻿/**
 * copyright (c) 2010 IxD Consultant (http://www.ixdc.net)
 * license
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License <http://www.gnu.org/copyleft/gpl.html> for more details.
   
 * version: 3.2
 * date: 19/03/2016
 * Actionscript 3.0
 * Flash Player 9
 * @author Ahmed Abbas, a.abbas@ixdc.net
 * @contributors
		Jan Jonas
			Leiter Softwareentwicklung
			learnbit Internet und Multimedia GmbH
			Buchwiese, Germany
		Yasmin
			aka geekgirl
 *
 * requirements:
 
 	1. dynamic TextField.
	2. HTML enabled TextField.
	3. pre-assigned TextFormat.
	4. Arabic fonts.
	
 * 
 * properties:
 
 	1. data: for referencing original input string.
	2. wrapFactor: optional reduction value (Mac only) for wrapping correction when using bi-directional text  (default = 0.9).
	3. htmlLines: array of spliced text block HTML lines.
	4. embedFonts: boolean (default = true).
	5. hindiNumeralsOnly: boolean (default = false).
 
 *
 * notes:
 
 	1. for embedding fonts, when using StyleSheet it must be assigned after calling this method and after setting TextFormat for proper font detection, for example:
	
		import net.ixdc.utils.StringUtils;
		var utils:StringUtils = new StringUtils();
		var format:TextFormat = new TextFormat();
		format.color = 0x0066FF;
		output.htmlText = utils.parseArabic("ARABIC TEXT", output.width, format);
		var styles:StyleSheet = new StyleSheet();
		styles.setStyle("a:hover", {color:"#000099"});
		output.stylesheet = styles;
	2. for matching/comparing input string, when referencing original input string use the data property (utils.data) instead of (TextField.text) where "utils" is an instance of StringUtils class.
	3. for bullet rendering, include the bullet character "•" in the Character Embedding menu.
 
 *
 * features supported:
 
 	1.  embedding fonts (just put a dynamic textfield on-stage and select at least Basic Latin (95 glyphs) and Arabic (1088 glyphs) from the Character Embedding menu).
	2.  Arabic ligatures.
	3.  Arabic diacritics.
	4.  word wrapping.
	5.  bi-directional text.
	6.  HTML text.
	7.  loading external text on run-time.
	8.  Windows/Mac/Linux support.
	9.  Arabic enabled input fields using actionscript method (createArabicInput) with the help of javascript (arabicinput.js).
	10. Urdu support.
 
 *
 * omitted features:
 
    1. latinOnly
    2. americanFormat
 
 *
 * example:
 	
	package {
		import flash.display.Sprite;
		import flash.events.Event;
		import flash.net.URLLoader;
		import flash.net.URLRequest;
		import flash.text.TextField;
		import flash.text.TextFieldAutoSize;
		import flash.text.TextFormat;
		import flash.events.MouseEvent;
		import net.ixdc.utils.StringUtils;
		public class ProperArabicTextAS3 extends Sprite {
			public var utils:StringUtils;
			public var format:TextFormat;
			public var output:TextField;
			public function ProperArabicTextAS3() {
				utils = new StringUtils();
				utils.wrapFactor = .8; // optional
				output = new TextField();
				output.width = stage.stageWidth-20;
				output.autoSize = TextFieldAutoSize.LEFT;
				output.embedFonts = true;
				output.wordWrap = true;
				output.multiline = true;
				addChild(output);
				format = new TextFormat();
				format.font = "Arial";
				format.size = 14;
				format.align = "right";
				format.rightMargin = 4;
				var xmlLoader:URLLoader = new URLLoader();
				xmlLoader.addEventListener(Event.COMPLETE, showXML);
				xmlLoader.load(new URLRequest("arabic.xml"));
			}
			public function showXML(event:Event):void {
				XML.ignoreWhitespace = true; 
				var arabic:XML = new XML(event.target.data);
				output.htmlText = utils.parseArabic(arabic.text(), output.width, format);
			}
		}
	}
	
	<?xml version="1.0" encoding="utf-8"?>
	<arabic>
		<![CDATA[الهدف من لعبة الكلمات المتقاطعة هو ملء المربعات البيضاء، وتشكيل الكلمات أو العبارات، عن طريق حل القرائن التي تؤدي إلى إجابات. والمربعات السوداء تستخدم لفصل الكلمات أو العبارات. ويوضع لكل صف أو عمود رقم، ثم يكتب أمام الرقم ما يشابه الكلمة المطلوب كتابتها في المربعات، وتحوي المجلات أنواعا من هذه اللعبة؛ فمنها سهل وآخر صعب، والكلمات المتقاطعة في مجلة معينة تتبع أسلوب يختلف عن الأخرى، تبعا لأسلوب اللغة المستعملة والبلد، وتقوم المجلات والصحف بوضع هذه اللعبة لتسلية القراء. ظهرت أول لعبة للكلمات المتقاطعة في صحيفة newyork world وذلك في 21 ديسمبر عام 1913، وأصبحت من الألعاب الرائجة في الولايات المتحدة، ومنها انتقلت إلى بقية دول العالم، وبمختلف اللغات، وكان أول من أدخلها إلى الصحافة هو آرثر وين.]]>
	</arabic>
 
 **/
package net.ixdc.utils { 
	import flash.external.ExternalInterface;
	import flash.geom.Point;
	import flash.system.Capabilities;
	import flash.text.TextField;
	import flash.text.TextFieldAutoSize;
	import flash.text.TextFormat;
	import flash.text.TextLineMetrics;
	import flash.display.MovieClip;
	public class StringUtils extends MovieClip {
		private var chars:String;
		private var specialChars:String = "";
		private var arabicPunctuation:String;
		private var isMAC:Boolean;
		public var data:String;
		public var wrapFactor:Number = .9; // optional reduction for wrapping correction when using bi-directional text, possible values: .9 or less
		public var htmlLines:Array;
		public var embedFonts:Boolean = true;
		public var latinOnly:Boolean; // omitted
		public var hindiNumeralsOnly:Boolean = false;
		public var americanFormat:Boolean; // omitted
		private var arabicDefaultRange:Boolean;
		private var arabicPresentationFormsARange:Boolean;
		private var arabicPresentationFormsBRange:Boolean;
		private var arabicExtendedARange:Boolean;
		private var arabicSupplementRange:Boolean;
		function StringUtils() {
			// check for OS, it seems like Flash Player behaves differently between Windows and Mac, especially for embedding fonts!
			isMAC = false;
			var version:String = Capabilities.version;
			if (version.toLowerCase().indexOf("mac") != -1) {
				isMAC = true;
			}
			specialChars += String.fromCharCode(1575, 1571, 1573, 1570, 1583, 1672, 1584, 1585, 1681, 1586, 1688, 1608, 1572, 1569);
			arabicPunctuation = String.fromCharCode(0x0609, 0x060A, 0x061B, 0x061E, 0x061F, 0x060C, 0x060D, 0x066A, 0x066B, 0x066C, 0x066D, 0x06D4);
		}
		/**
		 * @public
		 * creates proper arabic HTML string.
		 * 
		 * @param input String required value of original arabic string.
		 * @param lineWidth Number required value of wrapping line width.
		 * @param fieldFormat TextFormat optional value of final text format.
		 * @return String value with correct arabic text.
		 */
		public function parseArabic(input:String, lineWidth:Number, fieldFormat:TextFormat):String {
			var time:Date = new Date();
			var begin:Number = time.getTime();
			var end:Number;
			trace('StringUtils > parseArabic | AS3 v3.2\n');
			trace('embedFonts: '+embedFonts+'\n');
			 // referencing original input string to use for matching/comparing instead of (TextField.text)
			data = input;
			if (!fieldFormat) {
				fieldFormat = new TextFormat();
				fieldFormat.font = "Traditional Arabic";
				fieldFormat.size = 12;
				fieldFormat.align = "right";
				fieldFormat.rightMargin = 4;
			}
			htmlLines = [];
			var string:String = "";
			if (input.length>0) {
				chars = input;
				// format hard lines-breaks
				chars = chars.split("\r\n").join("\n");
				chars = chars.split("\r").join("\n");
				chars = chars.split("\n").join("<br>");
				// split bulleted list
				splitBulletList();
				// replace HTML opening and closing tags
				chars = chars.split("<").join("&lt;");
				chars = chars.split(">").join("&gt;");
				// convert numbers to hindi
				if (hindiNumeralsOnly) {
					chars = convertNumbers();
				}
				// remove ASCII control characters.
				chars = cleanString();
				// temporarily create an autosized textfield off-stage to use for proper right to left wrapping
				var temp:TextField = new TextField();
				temp.x = -1000;
				temp.y = -1000;
				temp.width = 10;
				temp.height = fieldFormat.size+4;
				addChild(temp);
				temp.autoSize = TextFieldAutoSize.LEFT;
				temp.embedFonts = true;
				temp.selectable = false;
				temp.text = "";
				var tempObj:Object;
				var tempData:String = "";
				var measureLine:Boolean;
				var arabicString:String = "";
				var htmlString:String = "";
				var defaultString:String = "";
				var chunks:Array = [];
				var code:Number;
				var toggleHTML:Boolean = false;
				var toggleArabic:Boolean = false;
				var toggleNumeric:Boolean = false;
				var openHTMLTag:Boolean = false;
				var closeHTMLTag:Boolean = false;
				var validNextChar:Boolean = false;
				var validArabicSeq:Boolean = false;
				var htmlLineBreaks:Array = [];
				var htmlTags:Object;
				var multiLines:Array = [];
				var tempLine:Array;
				var tempArray:Array;
				var tempString:String;
				var plainString:String;
				var i:Number = 0;
				var v:Number = 0;
				while (i<chars.length) {
					code = chars.charCodeAt(i);
					openHTMLTag = chars.charAt(i) == "&" && chars.charAt(i+1) == "l" && chars.charAt(i+2) == "t" && chars.charAt(i+3) == ";" && chars.charAt(i+4) != " ";
					if (openHTMLTag) {
						toggleHTML = true;
					}
					validArabicSeq = validateArabicChar(i) && !toggleHTML || validateSymbol(i) && !toggleHTML;
					if (validArabicSeq) {
						toggleArabic = true;
					} else {
						toggleArabic = false;
					}
					if (toggleArabic) {
						if (validateSymbol(i) && !validateNextArabicChar(i) && defaultString != "") {
							toggleArabic = false;
						}
					}
					if (toggleHTML) {
						if (arabicString != "") {
							chunks.push({arabic:true, html:false, value:arabicString});
							arabicString = "";
						}
						if (defaultString != "") {
							chunks.push({arabic:false, html:false, value:defaultString});
							defaultString = "";
						}
						htmlString += chars.charAt(i);
					} else if (toggleArabic) {
						if (defaultString != "") {
							chunks.push({arabic:false, html:false, value:defaultString});
							defaultString = "";
						}
						if (htmlString != "") {
							htmlString = htmlString.toLowerCase();
							if (htmlString.indexOf("&lt;br /&gt;") != -1) {
								htmlLineBreaks = htmlString.split("&lt;br /&gt;");
								htmlString = htmlLineBreaks.join("&lt;br&gt;");
							}
							if (htmlString.indexOf("&lt;br&gt;") != -1) {
								htmlLineBreaks = htmlString.split("&lt;br&gt;");
								htmlString = htmlLineBreaks.join("");
							}
							chunks.push({arabic:false, html:true, value:htmlString});
							htmlString = "";
						}
						arabicString += replaceArabicChar(i);
					} else {
						if (arabicString != "") {
							chunks.push({arabic:true, html:false, value:arabicString});
							arabicString = "";
						}
						if (htmlString != "") {
							htmlString = htmlString.toLowerCase();
							if (htmlString.indexOf("&lt;br /&gt;") != -1) {
								htmlLineBreaks = htmlString.split("&lt;br /&gt;");
								htmlString = htmlLineBreaks.join("&lt;br&gt;");
							}
							if (htmlString.indexOf("&lt;br&gt;") != -1) {
								htmlLineBreaks = htmlString.split("&lt;br&gt;");
								htmlString = htmlLineBreaks.join("");
							}
							chunks.push({arabic:false, html:true, value:htmlString});
							htmlString = "";
						}
						defaultString += chars.charAt(i);
					}
					if (!toggleHTML) {
						tempData += chars.charAt(i);
						// line break for proper wrapping
						if (htmlLineBreaks.length>0) {
							multiLines.push(chunks);
							chunks = [];
							tempData = "";
							if (htmlLineBreaks.length>=2) {
								for (v=0; v<htmlLineBreaks.length-2; v++) {
									multiLines.push([]);
								}
							}
							htmlLineBreaks = [];
						} else {
							temp.htmlText = tempData;
							temp.setTextFormat(fieldFormat);
							measureLine = false;
							if (isMAC) {
								var metrics:TextLineMetrics = temp.getLineMetrics(0);
								if (metrics.width>=(lineWidth*wrapFactor)-(fieldFormat.leftMargin ? Number(fieldFormat.leftMargin) : 0)-(fieldFormat.rightMargin ? Number(fieldFormat.rightMargin) : 0)) {
									measureLine = true;
								}
							} else {
								if (Math.ceil(temp.width)>=(lineWidth*wrapFactor)-(fieldFormat.leftMargin ? Number(fieldFormat.leftMargin) : 0)-(fieldFormat.rightMargin ? Number(fieldFormat.rightMargin) : 0)) {
									measureLine = true;
								}
							}
							if (measureLine && chars.charAt(i) == " ") {
								if (arabicString != "") {
									chunks.push({arabic:true, html:false, value:arabicString});
									arabicString = "";
								}
								if (defaultString != "") {
									chunks.push({arabic:false, html:false, value:defaultString});
									defaultString = "";
								}
								multiLines.push(chunks);
								chunks = [];
								tempData = "";
							}
						}
					}
					// closing HTML tags after grabbing all characters !
					closeHTMLTag = chars.charAt(i-3) == "&" && chars.charAt(i-2) == "g" && chars.charAt(i-1) == "t" && chars.charAt(i) == ";";
					if (closeHTMLTag) {
						toggleHTML = false;
					}
					i++;
				}
				// collect remaining chunks
				if (htmlString != "") {
					htmlString = htmlString.toLowerCase();
					if (htmlString.indexOf("&lt;br /&gt;") != -1) {
						tempArray = htmlString.split("&lt;br /&gt;");
						htmlString = tempArray.join("&lt;br&gt;");
					}
					if (htmlString.indexOf("&lt;br&gt;") != -1) {
						tempArray = htmlString.split("&lt;br&gt;");
						htmlString = tempArray.join("");
					}
					chunks.push({arabic:false, html:true, value:htmlString});
					htmlString = "";
				}
				if (arabicString != "") {
					chunks.push({arabic:true, html:false, value:arabicString});
				}
				if (defaultString != "") {
					chunks.push({arabic:false, html:false, value:defaultString});
				}
				if (chunks.length>0) {
					multiLines.push(chunks);
				}
				// remove temporary TextField
				removeChild(temp);
				// auto-complete HTML tags at line breaks
				tempArray = [];
				for (i=0; i<multiLines.length; i++) {
					htmlTags = getLastHTMLTag(multiLines[i]);
					if (!htmlTags.hasCloseTag) {
						multiLines[i].push({arabic:false, html:true, value:htmlTags.closeTag});
						tempArray.push({index:i+1, value:htmlTags.openTag});
					}
				}
				if (tempArray.length>0) {
					for (i=0; i<tempArray.length; i++) {
						multiLines[tempArray[i].index].splice(0, 0, {arabic:false, html:true, value:tempArray[i].value});
					}
				}
				// build output string
				for (i=0; i<multiLines.length; i++) {
					tempString = "";
					tempLine = multiLines[i];
					// reorder plain HTML chunks for proper rendering
					tempLine = reorderHTMLTags(multiLines[i]);
					for (v=0; v<tempLine.length; v++) {
						if (tempLine[v].arabic) {
							// reverse Arabic characters (for embedFonts only)
							if (embedFonts) {
								tempArray = tempLine[v].value.split("");
								tempArray.reverse();
								tempLine[v].value = tempArray.join("");
							}
						}
						if (tempLine[v].html) {
							// revert back HTML opening and closing tags
							tempLine[v].value = tempLine[v].value.split("&lt;").join("<");
							tempLine[v].value = tempLine[v].value.split("&gt;").join(">");
						}
						//trace('arabic: '+tempLine[v].arabic+' html: '+tempLine[v].html);
						//trace(tempLine[v].value);
						tempString += tempLine[v].value;
					}
					// reverse back numeric sequence
					plainString = "";
					toggleHTML = false;
					v = 0;
					while (v<tempString.length) {
						if (tempString.charAt(v) == "<" && tempString.charAt(v+1) != " ") {
							toggleHTML = true;
						}
						if (isNumeric(plainString+tempString.charAt(v))) {
							toggleNumeric = true;
						} else {
							toggleNumeric = false;
						}
						if (!toggleHTML && toggleNumeric) {
							plainString += tempString.charAt(v);
						}
						if (tempString.charAt(v) == ">") {
							toggleHTML = false;
						}
						v++;
					}
					if (plainString.indexOf(", ") != -1) {
						tempArray = plainString.split(", ");
						tempArray.reverse();
						tempString = tempString.split(plainString).join(tempArray.join(", "));
					}
					// flip brackets !
					tempObj = flipBracket(tempString);
					tempString = tempObj.line;
					// fix messed-up chevrons !
					tempString = tempString.split("؛lt&").join("&gt;");
					if (embedFonts) {
						tempString = tempString.split("؛gt&").join("&lt;");
					} else {
						tempString = tempString.split("&gt؛").join("&lt;");
						// trying to fix bullet position with bi-directional line (for none-embedFonts only) !
						if (tempString.charAt(1) == "•") {
							code = tempString.charCodeAt(3);
							arabicDefaultRange = code >= String.fromCharCode(0x0600).charCodeAt(0) && code <= String.fromCharCode(0x06FF).charCodeAt(0);
							arabicPresentationFormsARange = code >= String.fromCharCode(0xFB50).charCodeAt(0) && code <= String.fromCharCode(0xFDFF).charCodeAt(0);
							arabicPresentationFormsBRange = code >= String.fromCharCode(0xFE70).charCodeAt(0) && code <= String.fromCharCode(0xFEFF).charCodeAt(0);
							arabicExtendedARange = code >= String.fromCharCode(0x08A0).charCodeAt(0) && code <= String.fromCharCode(0x08FF).charCodeAt(0);
							arabicSupplementRange = code >= String.fromCharCode(0x0750).charCodeAt(0) && code <= String.fromCharCode(0x077F).charCodeAt(0);
							validNextChar = (arabicDefaultRange || arabicPresentationFormsARange || arabicPresentationFormsBRange || arabicExtendedARange || arabicSupplementRange);
							if (!validNextChar) {
								tempArray = tempString.split(" • ");
								tempArray.reverse();
								tempString = tempArray.join(" • ");
							}
						}
						// trying to fix single bracket occurance with bi-directional line (for none-embedFonts only) !
						if (tempObj.bracketsCount == 1) {
							tempArray = tempString.split(tempObj.flippedBracket);
							if (!validateArabicChunck(tempArray[0]) && !isNumeric(tempArray[0])) {
								tempArray.splice(1, 0, tempObj.flippedBracket);
								tempArray.splice(0, 0, String.fromCharCode(0xFE76)); // prepending an arabic character ?!
								tempString = tempArray.join("");
							}
						}
					}
					// save a copy of resulting line
					htmlLines.push(tempString);
					string += tempString+'<br>';
				}
				string = '<p align="right">'+string+'</p>';
			}
			// convert text format to html tags
			string = getHTMLFormat(string, fieldFormat);
			time = new Date();
			end = time.getTime();
			trace('parse complete in '+(end-begin)+' ms');
			trace('-------------------------------------------------\n');
			// returning proper arabic text
			return string;
		}
		/**
		 * @private
		 * helper method detects valid Arabic only characters.
		 * 
		 * @param i Number index of target character.
		 * @return Boolean value of valid Arabic only character.
		 */
		private function validateArabicChar(i:Number):Boolean {
			var valid:Boolean = false;
			var code:Number = chars.charCodeAt(i);
			arabicDefaultRange = code >= String.fromCharCode(0x0600).charCodeAt(0) && code <= String.fromCharCode(0x06FF).charCodeAt(0);
			arabicPresentationFormsARange = code >= String.fromCharCode(0xFB50).charCodeAt(0) && code <= String.fromCharCode(0xFDFF).charCodeAt(0);
			arabicPresentationFormsBRange = code >= String.fromCharCode(0xFE70).charCodeAt(0) && code <= String.fromCharCode(0xFEFF).charCodeAt(0);
			arabicExtendedARange = code >= String.fromCharCode(0x08A0).charCodeAt(0) && code <= String.fromCharCode(0x08FF).charCodeAt(0);
			arabicSupplementRange = code >= String.fromCharCode(0x0750).charCodeAt(0) && code <= String.fromCharCode(0x077F).charCodeAt(0);
			if (i>=0 && i<chars.length) {
				if (arabicDefaultRange || arabicPresentationFormsARange || arabicPresentationFormsBRange || arabicExtendedARange || arabicSupplementRange) {
					valid = true;
				}
			}
			return valid;
		}
		/**
		 * @private
		 * helper method detects valid Arabic only chunk.
		 * 
		 * @param input String index of target chunk.
		 * @return Boolean value of valid Arabic only chunk.
		 */
		private function validateArabicChunck(input:String):Boolean {
			var valid:Boolean = false;
			var code:Number = 0;
			var i:Number = 0;
			while (i<input.length) {
				code = input.charCodeAt(i);
				arabicDefaultRange = code >= String.fromCharCode(0x0600).charCodeAt(0) && code <= String.fromCharCode(0x06FF).charCodeAt(0);
				arabicPresentationFormsARange = code >= String.fromCharCode(0xFB50).charCodeAt(0) && code <= String.fromCharCode(0xFDFF).charCodeAt(0);
				arabicPresentationFormsBRange = code >= String.fromCharCode(0xFE70).charCodeAt(0) && code <= String.fromCharCode(0xFEFF).charCodeAt(0);
				arabicExtendedARange = code >= String.fromCharCode(0x08A0).charCodeAt(0) && code <= String.fromCharCode(0x08FF).charCodeAt(0);
				arabicSupplementRange = code >= String.fromCharCode(0x0750).charCodeAt(0) && code <= String.fromCharCode(0x077F).charCodeAt(0);
				if (arabicDefaultRange || arabicPresentationFormsARange || arabicPresentationFormsBRange || arabicExtendedARange || arabicSupplementRange) {
					valid = true;
					break;
				}
				i++;
			}
			return valid;
		}
		/**
		 * @private
		 * helper method detects next valid Arabic only characters.
		 * 
		 * @param i Number index of target character.
		 * @return Boolean value of next valid Arabic only character.
		 */
		private function validateNextArabicChar(i:Number):Boolean {
			var valid:Boolean = false;
			if (i>=0 && i<chars.length) {
				if (validateArabicChar(i)) {
					valid = true;
				}
				if (chars.charAt(i) == " " || validateSymbol(i)) {
					valid = validateNextArabicChar(i+1);
				}
			}
			return valid;
		}
		/**
		 * @private
		 * helper method detects previous valid Arabic only characters.
		 * 
		 * @param i Number index of target character.
		 * @return Boolean value of previous valid Arabic only character.
		 */
		private function validatePrevArabicChar(i:Number):Boolean {
			var valid:Boolean = false;
			if (i>=0 && i<chars.length) {
				if (validateArabicChar(i)) {
					valid = true;
				}
				if (chars.charAt(i) == " " || validateSymbol(i)) {
					valid = validatePrevArabicChar(i-1);
				}
			}
			return valid;
		}
		/**
		 * @private
		 * helper method detects valid symbols (including punctuation and excluding numerals).
		 * 
		 * @param i Number index of target character.
		 * @return Boolean value of valid symbols.
		 */
		private function validateSymbol(i:Number):Boolean {
			var valid:Boolean = false;
			var code:Number = chars.charCodeAt(i);
			if (code >= 32 && code <= 47 || code >= 58 && code <= 63 || code >= 91 && code <= 93 || code >= 123 && code <= 126 || code == 163 || code == 171 || code == 187 || code == 1548 || code == 8226 || code == 9642 || code == 10216 || code == 10217) {
				valid = true;
			}
			return valid;
		}
		/**
		 * @private
		 * helper method detects valid opening bracket.
		 * 
		 * @param i Number index of target character.
		 * @param src String value of search string.
		 * @return Boolean value of valid opening bracket.
		 */
		private function validateOpenBracket(i:Number, src:String):Boolean {
			var valid:Boolean = false;
			var code:Number = src.charCodeAt(i);
			if (code == 40 || code == 91 || code == 123 || code == 171 || code == 10216) {
				valid = true;
			}
			return valid;
		}
		/**
		 * @private
		 * helper method detects valid closing bracket.
		 * 
		 * @param i Number index of target character.
		 * @param src String value of search string.
		 * @return Boolean value of valid closing bracket.
		 */
		private function validateCloseBracket(i:Number, src:String):Boolean {
			var valid:Boolean = false;
			var code:Number = src.charCodeAt(i);
			if (code == 41 || code == 93 || code == 125 || code == 187 || code == 10217) {
				valid = true;
			}
			return valid;
		}
		/**
		 * @private
		 * helper method for proper brackets with bi-directional and text wrapping.
		 * 
		 * @param line String value of single text line.
		 * @return Object value of valid single line brackets with count.
		 */
		private function flipBracket(line:String):Object {
			var brackets:Array = [];
			var bracket:Object;
			var flippedBracket:String;
			var lineArray:Array;
			var i:Number = 0;
			while (i<line.length) {
				if (validateOpenBracket(i, line) || validateCloseBracket(i, line)) {
					brackets.push({index:i, code:line.charCodeAt(i)});
				}
				i++;
			}
			if (brackets.length>0) {
				for (i=0; i<brackets.length; i++) {
					bracket = brackets[i];
					switch (bracket.code) {
						case 40:
						flippedBracket = String.fromCharCode(41);
						break;
						case 91:
						flippedBracket = String.fromCharCode(93);
						break;
						case 123:
						flippedBracket = String.fromCharCode(125);
						break;
						case 171:
						flippedBracket = String.fromCharCode(187);
						break;
						case 10216:
						flippedBracket = String.fromCharCode(10217);
						break;
						case 41:
						flippedBracket = String.fromCharCode(40);
						break;
						case 93:
						flippedBracket = String.fromCharCode(91);
						break;
						case 125:
						flippedBracket = String.fromCharCode(123);
						break;
						case 187:
						flippedBracket = String.fromCharCode(171);
						break;
						case 10217:
						flippedBracket = String.fromCharCode(10216);
						break;
					}
					lineArray = line.split("");
					lineArray.splice(bracket.index, 1, flippedBracket);
					line = lineArray.join("");
				}
			}
			return {line:line, bracketsCount:brackets.length, flippedBracket:flippedBracket};
		}
		/**
		 * @private
		 * helper method splits HTML bulleted list.
		 */
		private function splitBulletList() {
			chars = chars.split("<ul>").join("");
			chars = chars.split("</ul>").join("<br>");
			chars = chars.split("<li>").join("<br> • ");
			chars = chars.split("</li>").join("");
		}
		/**
		 * @private
		 * helper method for proper HTML tags syntax.
		 * 
		 * @param line Array value of current line chunks.
		 * @return Object value of opening and closing HTML tags.
		 */
		private function getLastHTMLTag(line:Array):Object {
			var hasCloseTag:Boolean = true;
			var openTag:String = "";
			var closeTag:String = "";
			var tagArray:Array;
			var i:Number = line.length-1;
			if (line.length>0) {
				while (i>=0) {
					if (line[i].html) {
						if (line[i].value.indexOf("&lt;/") == -1) {
							openTag = line[i].value;
						}
						break;
					}
					i--;
				}
			}
			if (openTag != "") {
				hasCloseTag = false;
				tagArray = openTag.split(" ");
				if (tagArray.length>0) {
					i = tagArray.length-1;
					while (i>=0) {
						if (tagArray[i].indexOf("&lt;") != -1) {
							closeTag += "&lt;/"+tagArray[i].slice(tagArray[i].indexOf("&lt;")+4)+"&gt;";
							if (closeTag.indexOf("&gt;&gt;") != -1) {
								closeTag = closeTag.substr(0,-4);
							}
						}
						i--;
					}
				}
			}
			return {hasCloseTag:hasCloseTag, openTag:openTag, closeTag:closeTag};
		}
		/**
		 * @private
		 * helper method for proper text line order in bi-directional case.
		 * 
		 * @param line Array value of text line generated object.
		 * @return Array value of proper text line order.
		 */
		private function reorderTextLine(line:Array):Array {
			var plainIndices:Array = [];
			var plainChunks:Array = [];
			var i:Number;
			for (i=0; i<line.length; i++) {
				if (!line[i].html) {
					plainIndices.push(i);
					plainChunks.push(line[i]);
				}
			}
			plainChunks.reverse();
			for (i=0; i<plainIndices.length; i++) {
				line.splice(plainIndices[i], 1, plainChunks[i]);
			}
			return line;
		}
		/**
		 * @private
		 * helper method for proper HTML tags order.
		 * 
		 * @param line Array value of text line generated object.
		 * @return Array value of proper HTML tags order.
		 */
		private function reorderHTMLTags(line:Array):Array {
			var i:Number;
			var v:Number;
			if (embedFonts) {
				var hasArabic:Boolean = false;
				var htmlIndices:Array = [];
				var htmlChunks:Array = [];
				for (i=0; i<line.length; i++) {
					if (line[i].arabic) {
						hasArabic = true;
						break;
					}
				}
				if (hasArabic) {
					line.reverse();
				}
				for (i=0; i<line.length; i++) {
					if (line[i].html) {
						htmlIndices.push(i);
						htmlChunks.push(line[i]);
					}
				}
				if (hasArabic) {
					// reverse line html chunks (for embedFonts only)
					htmlChunks.reverse();
					for (i=0; i<htmlIndices.length; i++) {
						line.splice(htmlIndices[i], 1, htmlChunks[i]);
					}
					// reverse back mixed latin-html chunks  (for embedFonts only) !!
				}
			} else {
				var htmlComesFirst:Boolean = false;
				var plainComesFirst:Boolean = false;
				var toggleHTML:Boolean = false;
				var lineCompound:Array = [];
				var htmlCompound:Array = [];
				var plainCompound:Array = [];
				for (i=0; i<line.length; i++) {
					if (line[i].html || toggleHTML) {
						if (!plainComesFirst) {
							htmlComesFirst = true;
						}
						if (plainCompound.length>0) {
							lineCompound.push(plainCompound);
							plainCompound = [];
						}
						htmlCompound.push(line[i]);
					} else {
						plainComesFirst = true;
						if (htmlCompound.length>0) {
							lineCompound.push(htmlCompound);
							htmlCompound = [];
						}
						plainCompound.push(line[i]);
					}
					if (line[i].html) {
						if (toggleHTML) {
							toggleHTML = false;
						} else {
							toggleHTML = true;
						}
					}
				}
				if (htmlComesFirst) {
					if (htmlCompound.length>0) {
						lineCompound.push(htmlCompound);
						htmlCompound = [];
					}
					if (plainCompound.length>0) {
						lineCompound.push(plainCompound);
						plainCompound = [];
					}
				} else {
					if (plainCompound.length>0) {
						lineCompound.push(plainCompound);
						plainCompound = [];
					}
					if (htmlCompound.length>0) {
						lineCompound.push(htmlCompound);
						htmlCompound = [];
					}
				}
				lineCompound.reverse();
				line = [];
				for (i=0; i<lineCompound.length; i++) {
					for (v=0; v<lineCompound[i].length; v++) {
						line.push(lineCompound[i][v]);
					}
				}
			}
			return line;
		}
		/**
		 * @private
		 * uses unicode character codes to get proper arabic characters with various cases.
		 * 
		 * @param chars String containing original text.
		 * @param i Number index of target character.
		 * @return String value with correct arabic character.
		 */
		private function replaceArabicChar(i:Number):String {
			var string:String;
			switch (chars.charCodeAt(i)) {
				case 1575: //    ا
				string = setChar(i, String.fromCharCode(0x0627), String.fromCharCode(0x0627), String.fromCharCode(0xFE8E), String.fromCharCode(0xFE8E));
				break;
				case 1571: //    أ
				string = setChar(i, String.fromCharCode(0x0623), String.fromCharCode(0x0623), String.fromCharCode(0xFE84), String.fromCharCode(0xFE84));
				break;
				case 1573: //    إ
				string = setChar(i, String.fromCharCode(0x0625), String.fromCharCode(0x0625), String.fromCharCode(0xFE88), String.fromCharCode(0xFE88));
				break;
				case 1570: //    آ
				string = setChar(i, String.fromCharCode(0x0622), String.fromCharCode(0x0622), String.fromCharCode(0xFE82), String.fromCharCode(0xFE82));
				break;
				case 1576: //    ب
				string = setChar(i, String.fromCharCode(0x0628), String.fromCharCode(0xFE91), String.fromCharCode(0xFE92), String.fromCharCode(0xFE90));
				break;
				case 1662: //    پ
				string = setChar(i, String.fromCharCode(0x067E), String.fromCharCode(0xFB58), String.fromCharCode(0xFB59), String.fromCharCode(0xFB57));
				break;
				case 1578: //    ت
				string = setChar(i, String.fromCharCode(0x062A), String.fromCharCode(0xFE97), String.fromCharCode(0xFE98), String.fromCharCode(0xFE96));
				break;
				case 1657: //    ٹ
				string = setChar(i, String.fromCharCode(0x0679), String.fromCharCode(0xFB68), String.fromCharCode(0xFB69), String.fromCharCode(0xFB67));
				break;
				case 1579: //    ث
				string = setChar(i, String.fromCharCode(0x062B), String.fromCharCode(0xFE9B), String.fromCharCode(0xFE9C), String.fromCharCode(0xFE9A));
				break;
				case 1580: //    ج
				string = setChar(i, String.fromCharCode(0x062C), String.fromCharCode(0xFE9F), String.fromCharCode(0xFEA0), String.fromCharCode(0xFE9E));
				break;
				case 1670: //    چ
				string = setChar(i, String.fromCharCode(0x686), String.fromCharCode(0xFB7C), String.fromCharCode(0xFB7D), String.fromCharCode(0xFB7B));
				break;
				case 1581: //    ح
				string = setChar(i, String.fromCharCode(0x062D), String.fromCharCode(0xFEA3), String.fromCharCode(0xFEA4), String.fromCharCode(0xFEA2));
				break;
				case 1582: //    خ
				string = setChar(i, String.fromCharCode(0x062E), String.fromCharCode(0xFEA7), String.fromCharCode(0xFEA8), String.fromCharCode(0xFEA6));
				break;
				case 1583: //    د
				string = setChar(i, String.fromCharCode(0x062F), String.fromCharCode(0x062F), String.fromCharCode(0xFEAA), String.fromCharCode(0xFEAA));
				break;
				case 1672: //    ڈ
				string = setChar(i, String.fromCharCode(0x0688), String.fromCharCode(0xFB88), String.fromCharCode(0xFB89), String.fromCharCode(0xFB89));
				break;
				case 1584: //    ذ
				string = setChar(i, String.fromCharCode(0x0630), String.fromCharCode(0x0630), String.fromCharCode(0xFEAC), String.fromCharCode(0xFEAC));
				break;
				case 1585: //    ر
				string = setChar(i, String.fromCharCode(0x0631), String.fromCharCode(0x0631), String.fromCharCode(0xFEAE), String.fromCharCode(0xFEAE));
				break;
				case 1681: //    ڑ
				string = setChar(i, String.fromCharCode(0x0691), String.fromCharCode(0xFB8C), String.fromCharCode(0xFB8D), String.fromCharCode(0xFB8D));
				break;
				case 1586: //    ز
				string = setChar(i, String.fromCharCode(0x0632), String.fromCharCode(0x0632), String.fromCharCode(0xFEB0), String.fromCharCode(0xFEB0));
				break;
				case 1688: //    ژ
				string = setChar(i, String.fromCharCode(0x0698), String.fromCharCode(0xFB8A), String.fromCharCode(0xFB8B), String.fromCharCode(0xFB8B));
				break;
				case 1587: //    س
				string = setChar(i, String.fromCharCode(0x0633), String.fromCharCode(0xFEB3), String.fromCharCode(0xFEB4), String.fromCharCode(0xFEB2));
				break;
				case 1588: //    ش
				string = setChar(i, String.fromCharCode(0x0634), String.fromCharCode(0xFEB7), String.fromCharCode(0xFEB8), String.fromCharCode(0xFEB6));
				break;
				case 1589: //    ص
				string = setChar(i, String.fromCharCode(0x0635), String.fromCharCode(0xFEBB), String.fromCharCode(0xFEBC), String.fromCharCode(0xFEBA));
				break;
				case 1590: //    ض
				string = setChar(i, String.fromCharCode(0x0636), String.fromCharCode(0xFEBF), String.fromCharCode(0xFEC0), String.fromCharCode(0xFEBE));
				break;
				case 1591: //    ط
				string = setChar(i, String.fromCharCode(0x0637), String.fromCharCode(0xFEC3), String.fromCharCode(0xFEC4), String.fromCharCode(0xFEC2));
				break;
				case 1592: //    ظ
				string = setChar(i, String.fromCharCode(0x0638), String.fromCharCode(0xFEC7), String.fromCharCode(0xFEC8), String.fromCharCode(0xFEC6));
				break;
				case 1593: //    ع
				string = setChar(i, String.fromCharCode(0x0639), String.fromCharCode(0xFECB), String.fromCharCode(0xFECC), String.fromCharCode(0xFECA));
				break;
				case 1594: //    غ
				string = setChar(i, String.fromCharCode(0x063A), String.fromCharCode(0xFECF), String.fromCharCode(0xFED0), String.fromCharCode(0xFECE));
				break;
				case 1601: //    ف
				string = setChar(i, String.fromCharCode(0x0641), String.fromCharCode(0xFED3), String.fromCharCode(0xFED4), String.fromCharCode(0xFED2));
				break;
				case 1700: //    ڤ
				string = setChar(i, String.fromCharCode(0xFB6A), String.fromCharCode(0xFB6C), String.fromCharCode(0xFB6D), String.fromCharCode(0xFB6B));
				break;
				case 1602: //    ق
				string = setChar(i, String.fromCharCode(0x0642), String.fromCharCode(0xFED7), String.fromCharCode(0xFED8), String.fromCharCode(0xFED6));
				break;
				case 1603: //    ك
				string = setChar(i, String.fromCharCode(0x0643), String.fromCharCode(0xFEDB), String.fromCharCode(0xFEDC), String.fromCharCode(0xFEDA));
				break;
				case 1705: //    ک
				string = setChar(i, String.fromCharCode(0x06A9), String.fromCharCode(0xFB90), String.fromCharCode(0xFB91), String.fromCharCode(0xFB8F));
				break;
				case 1711: //    گ
				string = setChar(i, String.fromCharCode(0x06AF), String.fromCharCode(0xFB94), String.fromCharCode(0xFB95), String.fromCharCode(0xFB93));
				break;
				case 1604: //    ل
				string = setChar(i, String.fromCharCode(0x0644), String.fromCharCode(0xFEDF), String.fromCharCode(0xFEE0), String.fromCharCode(0xFEDE));
				break;
				case 1605: //    م
				string = setChar(i, String.fromCharCode(0x0645), String.fromCharCode(0xFEE3), String.fromCharCode(0xFEE4), String.fromCharCode(0xFEE2));
				break;
				case 1606: //    ن
				string = setChar(i, String.fromCharCode(0x0646), String.fromCharCode(0xFEE7), String.fromCharCode(0xFEE8), String.fromCharCode(0xFEE6));
				break;
				case 1722: //    ں
				string = setChar(i, String.fromCharCode(0x06BA), String.fromCharCode(0xFB9E), String.fromCharCode(0xFB9F), String.fromCharCode(0xFB9F));
				break;
				case 1723: //    ڻ
				string = setChar(i, String.fromCharCode(0x06BB), String.fromCharCode(0xFBA2), String.fromCharCode(0xFBA3), String.fromCharCode(0xFBA1));
				break;
				case 1607: //    ه
				string = setChar(i, String.fromCharCode(0x0647), String.fromCharCode(0xFEEB), String.fromCharCode(0xFEEC), String.fromCharCode(0xFEEA));
				break;
				case 1726: //    ھ
				string = setChar(i, String.fromCharCode(0xFBAA), String.fromCharCode(0xFBAC), String.fromCharCode(0xFBAD), String.fromCharCode(0xFBAB));
				break;
				case 1729: //    ہ
				string = setChar(i, String.fromCharCode(0xFBA6), String.fromCharCode(0xFBA8), String.fromCharCode(0xFBA9), String.fromCharCode(0xFBA7));
				break;
				case 1577: //    ة
				string = setChar(i, String.fromCharCode(0x0629), "", "", String.fromCharCode(0xFE94));
				break;
				case 1608: //    و
				string = setChar(i, String.fromCharCode(0x0648), String.fromCharCode(0x0648), String.fromCharCode(0xFEEE), String.fromCharCode(0xFEEE));
				break;
				case 1572: //    ؤ
				string = setChar(i, String.fromCharCode(0x0624), String.fromCharCode(0x0624), String.fromCharCode(0xFE86), String.fromCharCode(0xFE86));
				break;
				case 1609: //    ى
				string = setChar(i, String.fromCharCode(0x0649), String.fromCharCode(0x0649), String.fromCharCode(0xFEF0), String.fromCharCode(0xFEF0));
				break;
				case 1746: //    ے
				string = setChar(i, String.fromCharCode(0x06D2), String.fromCharCode(0xFBAE), String.fromCharCode(0xFBAF), String.fromCharCode(0xFBAF));
				break;
				case 1610: //    ي
				case 1740: // TODO: caught during urdu support!
				string = setChar(i, String.fromCharCode(0x064A), String.fromCharCode(0xFEF3), String.fromCharCode(0xFEF4), String.fromCharCode(0xFEF2));
				break;
				case 1574: //     ئ
				string = setChar(i, String.fromCharCode(0x0626), String.fromCharCode(0xFE8B), String.fromCharCode(0xFE8C), String.fromCharCode(0xFE8A));
				break;
				case 1747: //    ۓ
				string = setChar(i, String.fromCharCode(0x06D3), String.fromCharCode(0xFBB0), String.fromCharCode(0xFBB1), String.fromCharCode(0xFBB1));
				break;
				case 1569: //    ء
				string = String.fromCharCode(0x0621);
				break;
				case 1600: //    ـ
				string = String.fromCharCode(0x0640);
				break;
				case 1614: //    َ
				string = setChar(i, String.fromCharCode(0xFE76), String.fromCharCode(0xFE76), String.fromCharCode(0xFE77), String.fromCharCode(0xFE76));
				break;
				case 1611: //    ً
				string = setChar(i, String.fromCharCode(0xFE70), String.fromCharCode(0xFE70), String.fromCharCode(0xFE71), String.fromCharCode(0xFE70));
				break;
				case 1616: //    ِ
				string = setChar(i, String.fromCharCode(0xFE7A), String.fromCharCode(0xFE7A), String.fromCharCode(0xFE7B), String.fromCharCode(0xFE7A));
				break;
				case 1613: //    ٍ
				string = String.fromCharCode(0xFE74);
				break;
				case 1615: //    ُ
				string = setChar(i, String.fromCharCode(0xFE78), String.fromCharCode(0xFE78), String.fromCharCode(0xFE79), String.fromCharCode(0xFE78));
				break;
				case 1612: //    ٌ
				string = String.fromCharCode(0xFE72);
				break;
				case 1618: //    ْ
				string = setChar(i, String.fromCharCode(0xFE7E), String.fromCharCode(0xFE7E), String.fromCharCode(0xFE7F), String.fromCharCode(0xFE7E));
				break;
				case 1617: //    ّ
				string = setChar(i, String.fromCharCode(0xFE7C), String.fromCharCode(0xFE7C), String.fromCharCode(0xFE7F), String.fromCharCode(0xFE7C));
				break;
				case 63: //    ?
				string = String.fromCharCode(0x061F);
				break;
				case 44: //    ,
				string = String.fromCharCode(0x060C);
				break;
				case 59: //    ;
				string = String.fromCharCode(0x061B);
				break;
				case 37: //    %
				string = String.fromCharCode(0x066A);
				break;
				default:
				string = chars.charAt(i);
				break;
			}
			return string;
		}
		/**
		 * @private
		 * detects special cases for arabic ligatures.
		 * 
		 * @param chars String containing original text.
		 * @param i Number index of target character.
		 * @param solo String arabic character without joints.
		 * @param begin String arabic character with a trailing joint.
		 * @param middle String arabic character with both joint.
		 * @param end String arabic character with an initial joint.
		 * @return String value with correct arabic character.
		 */
		private function setChar(i:Number, solo:String, begin:String, middle:String, end:String):String {
			var string:String = "";
			// detect lam-alef (ﻻ) cases
			if (chars.charAt(i) == "ل" && chars.charAt(i+1) == "ا") {
				if (validateArabicChar(i-1) && specialChars.indexOf(chars.charAt(i-1)) == -1) {
					string = String.fromCharCode(0xFEFC);
				} else {
					string = String.fromCharCode(0xFEFB);
				}
				chars = chars.substring(0, i)+string+chars.substring(i+2, chars.length);
			} else if (chars.charAt(i) == "ل" && chars.charAt(i+1) == "أ") {
				if (validateArabicChar(i-1) && specialChars.indexOf(chars.charAt(i-1)) == -1) {
					string = String.fromCharCode(0xFEF8);
				} else {
					string = String.fromCharCode(0xFEF7);
				}
				chars = chars.substring(0, i)+string+chars.substring(i+2, chars.length);
			} else if (chars.charAt(i) == "ل" && chars.charAt(i+1) == "إ") {
				if (validateArabicChar(i-1) && specialChars.indexOf(chars.charAt(i-1)) == -1) {
					string = String.fromCharCode(0xFEFA);
				} else {
					string = String.fromCharCode(0xFEF9);
				}
				chars = chars.substring(0, i)+string+chars.substring(i+2, chars.length);
			} else if (chars.charAt(i) == "ل" && chars.charAt(i+1) == "آ") {
				if (validateArabicChar(i-1) && specialChars.indexOf(chars.charAt(i-1)) == -1) {
					string = String.fromCharCode(0xFEF6);
				} else {
					string = String.fromCharCode(0xFEF5);
				}
				chars = chars.substring(0, i)+string+chars.substring(i+2, chars.length);
			} else {
				// check for arabic character position in word (solo, begin, middle, end)
				var isDiacriticPrev:Boolean = chars.charCodeAt(i-1) >= 1611 && chars.charCodeAt(i-1) <= 1618;
				var isDiacriticNext:Boolean = chars.charCodeAt(i+1) >= 1611 && chars.charCodeAt(i+1) <= 1618;
				if (i == 0) {
					if (specialChars.indexOf(chars.charAt(i)) != -1 || !validateArabicChar(i+1)) {
						string = solo;
					} else {
						string = begin;
					}
				} else if (i == chars.length-1) {
					if (specialChars.indexOf(chars.charAt(i-1)) != -1 || !validateArabicChar(i-1)) {
						string = solo;
					} else {
						string = end;
					}
				} else if (validateArabicChar(i-1) && validateArabicChar(i+1)) {
					if (specialChars.indexOf(chars.charAt(i-1)) != -1) {
						if (specialChars.indexOf(chars.charAt(i)) != -1 || isDiacriticNext && !validateArabicChar(i+2)) {
							string = solo;
						} else {
							string = begin;
						}
					} else {
						if (specialChars.indexOf(chars.charAt(i)) != -1 || chars.charAt(i+1) == "ء" || chars.charAt(i) == "ة") {
							if (chars.charAt(i-1) != "ة") {
								if (isDiacriticPrev && specialChars.indexOf(chars.charAt(i-2)) != -1) {
									string = begin;
								} else {
									string = end;
								}
							} else {
								string = begin;
							}
						} else {
							if (chars.charAt(i-1) != "ة") {
								if (isDiacriticPrev && specialChars.indexOf(chars.charAt(i-2)) != -1) {
									string = begin;
								} else if (isDiacriticNext && !validateArabicChar(i+2)) {
									string = end;
								} else {
									if (arabicPunctuation.indexOf(chars.charAt(i+1)) != -1) {
										string = end;
									} else {
										string = middle;
									}
								}
							} else {
								string = begin;
							}
						}
					}
				} else {
					if (validateArabicChar(i-1) && !validateArabicChar(i+1)) {
						if (specialChars.indexOf(chars.charAt(i-1)) != -1 || isDiacriticPrev && specialChars.indexOf(chars.charAt(i-2)) != -1) {
							string = solo;
						} else {
							string = end;
						}
					} else if (!validateArabicChar(i-1) && validateArabicChar(i+1)) {
						if (specialChars.indexOf(chars.charAt(i)) != -1) {
							string = solo;
						} else {
							string = begin;
						}
					} else if (!validateArabicChar(i-1) && !validateArabicChar(i+1)) {
						string = solo;
					}
				}
			}
			return string;
		}
		/**
		 * @private
		 * remove invalid ASCII control characters.
		 *
		 * @return String containing clean characters.
		 */
		private function cleanString():String {
			var string:String = "";
			var i:Number = 0;
			while (i<chars.length) {
				// ASCII control characters charcode range 0-31 and charcode 127
				// Reference: http://en.wikipedia.org/wiki/ASCII#ASCII_control_characters
				if (chars.charCodeAt(i)>31 && chars.charCodeAt(i) != 127) {
					string += chars.charAt(i);
				}
				i++;
			}
			return string;
		}
		/**
		 * @private
		 * helper method converts TextFormat to HTML format tags.
		 * 
		 * @param input String value input string.
		 * @param format TextFormat value input text format.
		 * @return Strinf value of valid HTML format tags.
		 */
		private function getHTMLFormat(input:String, format:TextFormat):String {
			var htmlFormat:String = '<textformat leftmargin="'+((format.leftMargin) ? format.leftMargin : 0)+'" rightmargin="'+((format.rightMargin) ? format.rightMargin : 0)+'" leading="'+((format.leading) ? format.leading : 0)+'"><p align="'+((format.align) ? format.align : "left")+'"><font face="'+((format.font) ? format.font : "Arial")+'" color="'+((format.color) ? getProperColor(Number(format.color)) : "#000000")+'" size="'+((format.size) ? format.size : 12)+'" letterspacing="'+((format.letterSpacing) ? format.letterSpacing : 0)+'" kerning="'+((format.kerning) ? format.kerning : 0)+'">';
			if (format.bold) {
				htmlFormat += "<b>";
			}
			if (format.italic) {
				htmlFormat += "<i>";
			}
			if (format.underline) {
				htmlFormat += "<u>";
			}
			htmlFormat += input;
			if (format.underline) {
				htmlFormat += "</u>";
			}
			if (format.italic) {
				htmlFormat += "</i>";
			}
			if (format.bold) {
				htmlFormat += "</b>";
			}
			htmlFormat += "</font></p></textformat>";
			return htmlFormat;
		}
		/**
		 * @public
		 * creates proper arabic input field.
		 * 
		 * @param flashID String swf object id in HTML document.
		 * @param inputID String generated HTML input id.
		 * @param field TextField target text input in flash.
		 * @param color Number optional value for HTML input color.
		 */
		public function createArabicInput(flashID:String, inputID:String, field:TextField, color:Number, bgColor:Number):void {
			var c:String = "#000000";
			if (color) {
				c = getProperColor(color);
			}
			var b:String = "none";
			if (bgColor) {
				b = getProperColor(bgColor);
			}
			var format:TextFormat = field.getTextFormat();
			var point:Point = new Point(field.x, field.y);
			field.localToGlobal(point);
			var type:String = "singleline";
			if (field.multiline) {
				type = "multiline";
			}
			ExternalInterface.call("createArabicInput", flashID, inputID, field.text, point.x, point.y, field.width, field.height, c, b, type);
		}
		/**
		 * @public
		 * reveal all generated arabic input fields.
		 */
		public function showArabicInput():void {
			ExternalInterface.call("revealAllArabicInputs");
		}
		/**
		 * @private
		 * validate hindi numerals
		 * 
		 * @return Boolean valid value
		 */
		private function isNumeric(num:String):Boolean {
			return !isNaN(parseInt(num));
		}
		/**
		 * @private
		 * convert numeric characters
		 * 
		 * @return String converted string
		 */
		private function convertNumbers():String {
			var string:String = "";
			var openHTMLTag:Boolean = false;
			var closeHTMLTag:Boolean = false;
			var toggleHTML:Boolean = false;
			var tempNumber:Array = [];
			var i:Number = 0;
			while (i<chars.length) {
				openHTMLTag = chars.charAt(i) == "&" && chars.charAt(i+1) == "l" && chars.charAt(i+2) == "t" && chars.charAt(i+3) == ";" && chars.charAt(i+4) != " ";
				if (openHTMLTag) {
					toggleHTML = true;
				}
				if (!toggleHTML) {
					switch (chars.charAt(i)) {
						case "0":
						string += String.fromCharCode(0x0660);
						break;
						case "1":
						string += String.fromCharCode(0x0661);
						break;
						case "2":
						string += String.fromCharCode(0x0662);
						break;
						case "3":
						string += String.fromCharCode(0x0663);
						break;
						case "4":
						string += String.fromCharCode(0x0664);
						break;
						case "5":
						string += String.fromCharCode(0x0665);
						break;
						case "6":
						string += String.fromCharCode(0x0666);
						break;
						case "7":
						string += String.fromCharCode(0x0667);
						break;
						case "8":
						string += String.fromCharCode(0x0668);
						break;
						case "9":
						string += String.fromCharCode(0x0669);
						break;
						default:
						string += chars.charAt(i);
						break;
					}
				} else {
					string += chars.charAt(i);
				}
				closeHTMLTag = chars.charAt(i-3) == "&" && chars.charAt(i-2) == "g" && chars.charAt(i-1) == "t" && chars.charAt(i) == ";";
				if (closeHTMLTag) {
					toggleHTML = false;
				}
				i++;
			}
			return string;
		}
		/**
		 * @private
		 * convert number with american standard ###,###,###.####
		 * 
		 * @return String formatted number
		 */
		private function toAmericanFormat(number:String):String {
			var str:String = "";
			var inc:Number = 0;
			for (var i:Number = number.length-1; i>=0; i--) {
				inc++;
				str += number.charAt(i);
				if (inc == 3) {
					inc = 0;
					str += ",";
				}
			}
			var arr:Array = str.split("");
			arr.reverse();
			str = arr.join("");
			return str;
		}
		/**
		 * @private
		 * helper method converts RGB color to HTML hexadecimal color.
		 * 
		 * @param input Number value of RGB color.
		 * @return String value of valid HTML hexadecimal color.
		 */
		private function getProperColor(input:Number):String {
			var color:String = "#";
			var checker:Number = 0;
			if (input.toString(16).length<6) {
				while (checker<6-input.toString(16).length) {
					color += "0";
					checker++;
				}
			}
			color += input.toString(16);
			return color;
		}
	}
}